# Copyright 2023 FBK

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
import torch
from torch import LongTensor

from pangolinn.seq2seq.base_tester import BaseTester


class EncoderPaddingTestCase(BaseTester):
    """
    This class provides unit tests to enforce that the module to be tested properly handles
    padding, i.e. returns the same results regardless of the amount of padding present in
    the batched input sequences.

    To use it to test your network:

     1. create a `PangolinnSeq2SeqModuleWrapper` that wraps your module (e.g. `MyWrapper`);
     2. create test class that extends `EncoderPaddingTestCase`;
     3. in your test class, override the class attribute `module_wrapper_class` by setting it to
        the **class** of your wrapper (e.g., `module_wrapper_class = MyWrapper`);
    """
    def setUp(self) -> None:
        self._wrapper_setup(EncoderPaddingTestCase)

    def test_padding_area_is_zero(self):
        """
        Tests that the padding area of the output contains all zeroes.
        Although the presence of non-zero elements in the passing area is not an issue
        on its own, elaborations (e.g., convolutions) on top of non-zero-padded tensors
        might cause issues.
        """
        # test both with multipliers of 4 and not, as many systems
        # may have a 2x or 4x downsampling factor, so we test both when
        # seq_len is divisible or not by the downsampling factor
        for max_batch_seq_len, shorter_seq_len in [(27, 13), (24, 16)]:
            rand_batch = self._rand_tensor(
                (2, max_batch_seq_len, self.module_wrapper.num_input_channels),
                self.module_wrapper.input_dtype)
            rand_batch[1, shorter_seq_len:, :] = 0
            batch_lens = torch.LongTensor([max_batch_seq_len, shorter_seq_len])
            expected_shape = [
                2,
                self.module_wrapper.output_sequence_length(max_batch_seq_len),
                self.module_wrapper.num_output_channels]
            output = self._forward_with_expected_shape(rand_batch, batch_lens, expected_shape)
            expected_shorter_seq_len = self.module_wrapper.output_sequence_length(shorter_seq_len)
            padding_area = output[1, expected_shorter_seq_len:, :]
            self.assertGreater(padding_area.numel(), 0)
            self.assertTrue(torch.all(padding_area == 0), f"non-zero entries in {padding_area}")

    def test_batch_size_does_not_matter(self):
        """
        Tests that for the same input we get the same output regardless of the amount of padding.
        """
        for max_batch_seq_len, shorter_seq_len in [(27, 13), (24, 16)]:
            rand_batch = self._rand_tensor(
                (5, max_batch_seq_len, self.module_wrapper.num_input_channels),
                self.module_wrapper.input_dtype)
            # multiple padded elements of same len
            for b_idx in [1, 2, 3]:
                rand_batch[b_idx, shorter_seq_len:, :] = 0
            # sequence of len 1
            rand_batch[4, 1:, :] = 0
            batch_lens = torch.LongTensor([
                max_batch_seq_len, shorter_seq_len, shorter_seq_len, shorter_seq_len, 1])
            expected_shape = [
                5,
                self.module_wrapper.output_sequence_length(max_batch_seq_len),
                self.module_wrapper.num_output_channels]
            output = self._forward_with_expected_shape(rand_batch, batch_lens, expected_shape)
            for i in range(5):
                item_len = batch_lens[i].item()
                item_valid_tokens = rand_batch[i, :item_len, :].unsqueeze(0)
                output_wo_padding = self.module_wrapper.forward(
                    item_valid_tokens, LongTensor([item_len]))
                item_out_len = self.module_wrapper.output_sequence_length(item_len)
                torch.testing.assert_close(
                        output[i, :item_out_len, :],
                        output_wo_padding[0, :, :])



# Copyright 2023 FBK

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
import torch

from pangolinn.seq2seq.base_tester import BaseTester


class CausalTestCase(BaseTester):
    """
    This class provides unit tests to enforce that the module to be tested is safe to
    be used in causal models, i.e. that do not look at future elements. This property
    should be enforced for e.g. Transformer autoregressive decoders.

    To use it to test your network:

     1. create a `PangolinnSeq2SeqModuleWrapper` that wraps your module (e.g. `MyWrapper`);
     2. create test class that extends `CausalTestCase`;
     3. in your test class, override the class attribute `module_wrapper_class` by setting it to
        the **class** of your wrapper (e.g., `module_wrapper_class = MyWrapper`);
    """
    def setUp(self) -> None:
        self._wrapper_setup(CausalTestCase)

    def test_gradient_not_flowing_from_future(self):
        """
        Checks that the gradient is not backpropagated to future input time steps, which should not
        be used to compute the output.
        """
        x = self._rand_tensor(
            (1, 10, self.module_wrapper.num_input_channels), self.module_wrapper.input_dtype)
        x.requires_grad = True
        expected_shape = [
            1,
            self.module_wrapper.output_sequence_length(10),
            self.module_wrapper.num_output_channels]
        y = self._forward_with_expected_shape(x, torch.LongTensor([10]), expected_shape).abs()
        for i in range(1, 9):
            # how many of the output elements depend on a prefix made of i elements
            y_prefix_end = self.module_wrapper.output_sequence_length(i)
            grad = torch.autograd.grad(y[:, y_prefix_end, :].sum(), x, retain_graph=True)[0]
            # Checks that the gradient for the prefix of x made of i elements is not zero,
            # while it is zero for the elements after the prefix of i elements, as the gradient
            # should not propagate to the future
            self.assertGreater(grad[0, :i, :].sum().abs(), 0.0)
            self.assertAlmostEqual(grad[0, i+1:, :].sum().item(), 0.0)

    def test_not_looking_at_the_future(self):
        """
        Tests that the module masks future elements and it does not look at them.
        """
        test_len = 20
        x = self._rand_tensor(
            (5, test_len, self.module_wrapper.num_input_channels), self.module_wrapper.input_dtype)
        batch_lens = torch.LongTensor([test_len] * 5)
        expected_shape = [
            5,
            self.module_wrapper.output_sequence_length(test_len),
            self.module_wrapper.num_output_channels]
        output = self._forward_with_expected_shape(x, batch_lens, expected_shape)
        for j in range(1, 19):
            # Checks that for each of the 20 elements we obtain the same prefix in the
            # results when feeding the model with the full input sequences and the input
            # prefix truncated at that element.
            partial_lens = torch.LongTensor([j] * 5)
            partial_output = self.module_wrapper.forward(x[:, :j, :], partial_lens)
            partial_output_len = partial_output.shape[1]
            torch.testing.assert_close(
                partial_output,
                output[:, :partial_output_len, :])



# Copyright 2023 FBK

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
import torch
from torch import nn, LongTensor, Tensor


class PangolinnSeq2SeqModuleWrapper:
    """
    Wrapper of your module used in pangolinn tests. To test your network, extend
    this class by implementing at least:

     - build_module;
     - num_input_channels;
     - forward.

    Depending on the behavior of your module you might need to override other methods.
    Please refer to the documentation of each method to check if you need (or not) to override it.
    """
    def __init__(self):
        self._module = self.build_module()
        self._module.eval()

    def build_module(self) -> nn.Module:
        """
        This method is responsible for building the module that you want to test.
        Override this method so that it creates an instance of the module to be tested.
        The module does not need to be initialized with specific weights.

        :return: the network you want to test.
        """
        raise NotImplementedError(
            "Please implement build_module to return the module to test")

    @property
    def num_input_channels(self) -> int:
        """
        This property has to be overridden and returns the number of channels expected in
        input by the module to be tested. It is typically the third dimension of
        sequence-to-sequence modules, in addition to the batch size and the sequence length.
        In the case of modules that expect a single integer (e.g., the text token id in text
        processing), this should be set to 1 and when overriding the forward method you can
        squeeze the last dimension before feeding the module with it.

        :return: the number of channels expected in the input tensor by the module to be tested.
        """
        raise NotImplementedError(
            "Please implement num_input_channels to return the number of channels expected by "
            "the module to be tested")

    def forward(self, x: Tensor, lengths: LongTensor) -> Tensor:
        """
        Processes `x` with the wrapped module and returns the output.

        :param x: the tensor to be fed to the wrapped module with shape (batch, seq_len, channels)
        :param lengths: tensor of shape (batch, ) that contains the length of the valid tokens
                        for each of the sequences in the batch.
        :return: the tensor produced by the module with shape (batch, seq_len, channels)
        """
        raise NotImplementedError(
            "Please implement forward to return the output of the wrapped module to be tested")

    @property
    def sequence_downsampling_factor(self) -> int:
        """
        If the module to be tested reduces the sequence length (e.g., as strided convolutions do),
        this property states the downsampling factor.

        :return: the downsampling factor over the sequence length (e.g., in speech processing
                 Transformer-based architectures, this is often 4). By default, this is set to 1,
                 which means no downsampling.
        """
        return 1

    @property
    def num_output_channels(self) -> int:
        """
        By default, this property returns as output channels the same as the input channels.
        If your module changes the number of channels, override this method accordingly.

        :return: the number of channels produced in output by the module to be tested.
        """
        return self.num_input_channels

    def output_sequence_length(self, input_sequence_len: int) -> int:
        """
        By default, this property returns as output sequence length the `input_sequence_len`
        divided by the downsampling factor. If your module has a more complicated function
        to determine the output sequence length from the input sequence length, override
        this method accordingly.

        :param input_sequence_len: the length of a sequence to be fed to the module.
        :return: the length of the valid tokens in the output sequence.
        """
        return ((input_sequence_len - 1) // self.sequence_downsampling_factor) + 1

    @property
    def input_dtype(self) -> torch.dtype:
        """
        :return: the dtype of the input tensor expected by the module. Defaults to torch.float.
        """
        return torch.float

    @property
    def max_value_allowed(self) -> int:
        """
        :return: if `input_dtype` is set to torch.int or another integer, this determines
                 the maximum value supported as input of the module to be tested. Defaults to 10.
        """
        return 10



# Copyright 2023 FBK

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
__all__ = ["CausalTestCase", "EncoderPaddingTestCase", "PangolinnSeq2SeqModuleWrapper"]

from .causal_tester import CausalTestCase  # noqa: F401
from .padding_tester import EncoderPaddingTestCase  # noqa: F401
from .seq2seq_module_wrapper import PangolinnSeq2SeqModuleWrapper  # noqa: F401



# Copyright 2023 FBK

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
import unittest
from typing import List, Tuple, Type

import torch
from torch import Tensor, LongTensor

from pangolinn.seq2seq.seq2seq_module_wrapper import PangolinnSeq2SeqModuleWrapper


class BaseTester(unittest.TestCase):
    """
    This class provides basic functions useful for pangolinn sequence-to-sequence testers.
    """
    module_wrapper_class: PangolinnSeq2SeqModuleWrapper.__class__

    def _wrapper_setup(self, pangolinn_class: Type):
        assert self.__class__ is not pangolinn_class, \
            "If you are seeing this error, most likely you have imported " \
            f"{pangolinn_class.__name__}. Instead, import pangolinn.seq2seq and " \
            f"extend seq2seq.{pangolinn_class.__name__}."
        assert self.module_wrapper_class is not None, \
            "Override the class attribute `module_wrapper_class` by setting it to the class of " \
            "your wrapper (e.g., `module_wrapper_class = MyWrapper`)."
        self.module_wrapper: PangolinnSeq2SeqModuleWrapper = self.module_wrapper_class()

    def _forward_with_expected_shape(
            self, x: Tensor, lengths: LongTensor, expected_shape: List[int]) -> Tensor:
        """
        :param x: tensor used as input of the module to be tested
        :param lengths: tensor containing the lengths of each sequence in `x`
        :param expected_shape: list of the expected dimensions
        :return: the output of the forward over the module to be tested
        """
        output = self.module_wrapper.forward(x, lengths)
        self.assertListEqual(
            expected_shape,
            list(output.size()),
            msg=f"Unexpected output shape {output.size()}. Model wrapper should return "
                "a tensor of shape (batch, seq_len, channels), with expected shape "
                f"{expected_shape}.")
        return output

    def _rand_tensor(self, shape: Tuple[int, int, int], dtype: torch.dtype) -> Tensor:
        if dtype.is_floating_point or dtype.is_complex:
            return torch.rand(shape, dtype=dtype)
        else:
            return torch.randint(self.module_wrapper.max_value_allowed, shape, dtype=dtype)
